/* Copyright (C) 2020-2021 BaseALT Ltd
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <QThread>

#include "dialog.h"

Dialog::Dialog(Process *process, QWidget *parent)
    : QDialog(parent),
      m_process(process),
      ui(new Ui::Dialog)
{
    m_cancelled = false;
    ui->setupUi(this);

    setModal(true);
    connect(ui->answers, &QDialogButtonBox::clicked,
            this, &Dialog::onButtonBoxClicked);
    connect(ui->box, &QDialogButtonBox::clicked,
            this, &Dialog::cancelOperation);

    connect(m_process, &Process::readyReadStandardOutput,
            this, &Dialog::onReadyReadStandardOutput);
    connect(m_process, &Process::readyReadStandardError,
            this, &Dialog::onReadyReadStandardError);
    connect(ui->accept, &QPushButton::clicked,
            this, &Dialog::acceptPassword);
    connect(ui->password, &QLineEdit::textChanged,
            this, &Dialog::onPasswordChanged);
    ui->password->hide();
    ui->accept->hide();
    ui->accept->setEnabled(false);
    ui->box->hide();

    m_first = true;
    ui->textEdit->setVisible(false);
    ui->progressBar->setVisible(false);
    adjustSize();
}

void Dialog::onPasswordChanged()
{
    ui->accept->setEnabled(!ui->password->text().isEmpty());
}

void Dialog::acceptPassword()
{
    m_first = false;
    m_process->write(ui->password->text().toUtf8() + '\n');
}

void Dialog::onReadyReadStandardError()
{
    m_process->readStandardError();
    QByteArray error = m_process->standardError();
    if (error.indexOf("Введите пароль:") != -1 ||
        error.indexOf("Type password:") != -1) {
        ui->password->show();
        ui->accept->show();
        ui->answers->hide();
    }
    if (!m_first)
        ui->textEdit->append(tr("Invalid password"));
}

void Dialog::onReadyReadStandardOutput()
{
    QByteArray output = m_process->standardOutput();
    ui->textEdit->setVisible(!output.isEmpty());

    if (output.contains("%\b\b\b\b")) {
        ui->progressBar->setVisible(true);
        ui->password->setVisible(false);
        ui->answers->setVisible(false);
        ui->accept->setVisible(false);
        ui->box->setVisible(true);
        auto maybeProgress = QString(output).split(' ', QString::SkipEmptyParts);
        for (auto i = maybeProgress.constBegin(); i != maybeProgress.constEnd(); ++i) {
            if (QString::number(ui->progressBar->value()) + "%\b\b\b\b" != (*i)) {
                int number = i->section('%', 0, 0).toInt();

                ui->progressBar->setValue(number);
                break;
            }
        }
    } else {
        ui->textEdit->setText(output);
    }
    question();
}

void Dialog::question()
{
    QDialogButtonBox::StandardButtons buttons = QDialogButtonBox::StandardButtons();
    auto keys = m_process->questionKeys(m_process->standardOutput());
    for (auto i = keys.constBegin();
         i != keys.constEnd();
         ++i) {
        Qt::Key key = (*i);
        if (key == Qt::Key_Y)
            buttons |= QDialogButtonBox::Yes;
        if (key == Qt::Key_N)
            buttons |= QDialogButtonBox::No;
        if (key == Qt::Key_C)
            buttons |= QDialogButtonBox::Cancel;
    }

    ui->answers->setStandardButtons(buttons);
}

void Dialog::onButtonBoxClicked(QAbstractButton *button)
{
    QDialogButtonBox::StandardButton key = ui->answers->standardButton(button);
    switch (key) {
    case QDialogButtonBox::Yes:
        m_process->write(QByteArray(1, 'Y'));
        ui->answers->setStandardButtons(QDialogButtonBox::StandardButtons());
        break;
    case QDialogButtonBox::No:
        m_process->write(QByteArray(1, 'N'));
        ui->answers->setStandardButtons(QDialogButtonBox::StandardButtons());
        break;
    case QDialogButtonBox::Cancel:
        m_process->write(QByteArray(1, 'C'));
        ui->answers->setStandardButtons(QDialogButtonBox::StandardButtons());
        break;
    default:
        break;
    }
}

void Dialog::cancelOperation()
{
    m_cancelled = true;
    m_process->terminate();
}

void Dialog::closeEvent(QCloseEvent *e)
{
    Q_UNUSED(e);
    if (m_process->state() == QProcess::Running)
        cancelOperation();
}

bool Dialog::cancelled()
{
    return m_cancelled;
}
