/* Copyright (C) 2020-2021 BaseALT Ltd
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <QDateTime>
#include <QFileDialog>
#include <QMessageBox>
#include <QResource>
#include <QStandardItemModel>
#include <QDesktopServices>
#include <QDebug>

#include "cache.h"
#include "process.h"
#include "mainwindow.h"
#include "utils.h"
#include "dialog.h"

MainWindow::MainWindow(QWidget *parent):
    QWidget(parent),
    ui(new Ui::MainWindow)
{
    ui->setupUi(this);
    loadStyleSheet();

    // global
    connect(ui->docView, &QPushButton::clicked,
            this, &MainWindow::viewDocument);
    connect(ui->docSelect, &QPushButton::clicked,
            this, &MainWindow::selectDocument);

    // verify
    connect(ui->sigSelectChk, &QPushButton::clicked,
            this, &MainWindow::selectSignature);
    connect(ui->sigChk, &QPushButton::clicked,
            this, &MainWindow::verifySignature);

    // sign
    connect(ui->generateName, &QPushButton::clicked,
            this, &MainWindow::generateName);
    connect(ui->sign, &QPushButton::clicked,
            this, &MainWindow::sign);
    connect(ui->signCompress, &QPushButton::clicked,
            this, [=](){
        m_compress = true;
        sign();
    });
    connect(ui->attached, &QCheckBox::stateChanged,
            this, &MainWindow::validateSignatureActions);

    connect(ui->extension, &QComboBox::currentTextChanged,
            this, &MainWindow::updateSignatureExtension);

    m_compress = false;

    m_dir = QDir::currentPath();
    m_cache.reset(new Cache());
    m_cache->createDir();

    m_waitTimer.setSingleShot(true);
    m_waitTimer.setInterval(2000);
    connect(&m_waitTimer, &QTimer::timeout,
            this, &MainWindow::showDialog);

    auto model = new QStandardItemModel();
    ui->certsTable->setModel(model);

    model->setHorizontalHeaderItem(0, new QStandardItem(tr("Subject")));
    model->setHorizontalHeaderItem(1, new QStandardItem(tr("Issuer")));
    model->setHorizontalHeaderItem(2, new QStandardItem(tr("SHA1 Fingerprint")));

    QItemSelectionModel *selectionModel = ui->certsTable->selectionModel();
    connect(selectionModel, &QItemSelectionModel::selectionChanged,
            this, &MainWindow::validateSignatureActions);

    qWarning() << "System architecture:" << QSysInfo::currentCpuArchitecture();

    if (!Utils::prepareCerts()) {
        disableControls();
        ui->tabWidget->setCurrentIndex(1);
        ui->result->setText(tr("Unable to find containers."));
    } else {
        loadCerts();
        validateSignatureActions();
    }
    ui->docView->setEnabled(false);
}

MainWindow::~MainWindow()
{
}

void MainWindow::init(QStringList files)
{
    if (files.isEmpty())
        return;

    QStringList documents;
    QStringList signatures;

    for (auto i = files.constBegin();
         i != files.constEnd();
         ++i)
    {
        QString file = (*i);
        QString extension = Utils::splitextSuffix(file);

        if (currentExtensions().contains(extension)) {
            signatures << file;
        } else {
            documents << file;
        }
    }

    if (!documents.isEmpty()) {
        changeDocument(documents.at(0));
    }
    if (!signatures.isEmpty()) {
        if (documents.isEmpty()) {
            QString document = Utils::splitextPrefix(signatures.at(0)).mid(18);
            if (QFile::exists(document))
                    changeDocument(document);

            changeSignatureFile(signatures.at(0));
        } else {
            changeSignatureFile(signatures.at(0));
        }
    }
}

bool MainWindow::checkCurrentDocumentExists()
{
    return (m_documentPath != "") && Utils::dirExists(Utils::splitDirname(m_documentPath));
}

QString MainWindow::getDocumentDir()
{
    if (checkCurrentDocumentExists()) {
        return Utils::splitDirname(m_documentPath);
    } else {
        return m_dir;
    }
}

QString MainWindow::getSignatureDir()
{
    if (checkCurrentDocumentExists()) {
        return Utils::splitDirname(m_documentPath);
    } else {
        return m_dir;
    }
}

void MainWindow::saveDir()
{
    if (checkCurrentDocumentExists()) {
        m_dir = Utils::splitDirname(m_documentPath);
    }
}

QStringList MainWindow::getTableData(int column)
{
    QStringList fingerprints;
    auto rows = ui->certsTable->selectionModel()->selectedRows(column);
    for (auto i = rows.constBegin();
         i != rows.constEnd();
         ++i)
    {
        fingerprints << i->data().toString();
    }
    return fingerprints;
}

void MainWindow::viewDocument()
{
    saveDir();

    QDesktopServices::openUrl(QUrl::fromLocalFile(m_documentPath));
}

void MainWindow::selectDocument()
{
    QFileDialog dialog(this, tr("Select document or archive", "Header"), getDocumentDir());
    if (dialog.exec())
        changeDocument(dialog.selectedFiles().at(0));
}

void MainWindow::selectSignature()
{
    QFileDialog dialog(this, tr("Select signature"), getSignatureDir());
    QStringList extensions = currentExtensions();
    extensions.replaceInStrings(".", "*.");
    QString formattedExtensions = extensions.join(" ");

    dialog.setNameFilters(QStringList()
                          << tr("Signatures (%1)").arg(formattedExtensions)
                          << tr("All files (*)"));
    if (dialog.exec())
        changeSignatureFile(dialog.selectedFiles()[0]);
}

void MainWindow::verifySignature()
{
    saveDir();

    Process *verifyProcess = new Process();
    connect(verifyProcess, &Process::completed,
            this, &MainWindow::onVerifyerCompleted);
    startWaiting(new Dialog(verifyProcess, this));
    verifyProcess->verify(m_documentPath, m_signaturePath);
}

void MainWindow::sign()
{
    QString dir = Utils::splitDirname(m_documentPath);
    QString file = ui->sigFilename->text();
    QString ext = Utils::splitextSuffix(file);
    QString path = QStringLiteral("%1/%2").arg(dir, file);
    if (m_fprintsPending.isEmpty()) {
        m_fprintsPending = getTableData(2);
        m_signaturesPending = QStringList();
    }

    QString insertedPath = path;
    if (getTableData(2).count() > 1) {
        insertedPath = Utils::insertInfix(path, QString::number(getTableData(2).indexOf(m_fprintsPending.at(0))));
    }

    addExtension(ext);
    saveDir();

    Process *process = new Process();
    connect(process, &Process::completed,
            this, &MainWindow::onSignerCompleted);

    if (!Utils::packageInstalled("cprocsp-rdr-gui-gtk") &&
        !Utils::packageInstalled("cprocsp-rdr-gui-gtk-64")) {
        startWaiting(new Dialog(process, this));
    }
    m_pendingSignature = insertedPath;
    if (!m_fprintsPending.isEmpty()) {
        QString sig = m_fprintsPending.first();
        process->sign(m_documentPath, insertedPath, sig,
                      ui->encoding->currentText(),
                      ui->attached->isChecked());
    }
}

void MainWindow::addExtension(QString extension)
{
    if (extension.isEmpty())
        return;

    if (!currentExtensions().contains(extension))
        ui->extension->addItem(extension);
    ui->extension->setCurrentText(extension);
}

bool MainWindow::processCompleted(QObject *p, Result *result)
{
    Process *process = qobject_cast<Process *>(p);

    if (!process)
        return false;

    bool isCancelled = false;
    if (m_dialog)
        isCancelled = m_dialog->cancelled();
    stopWaiting();

    if (isCancelled)
        return false;

    QString description;
    if (m_pending == PendingAction::Sign)
        description = tr("File can't be signed");
    else
        description = tr("File can't be verified");

    if (result->error() != Process::UnknownError) {
        error(tr("Error"), description, process->errorString());
        return false;
    }

    if (result->exitStatus() == Process::CrashExit) {
        error(tr("Crash"), description, process->errorString());
        return false;
    }

    Status status;
    if (process->errorOccured())
        status = Status::Error;
    else
        status = Status::Normal;

    showResult(process->standardOutput(), status);
    return status == Status::Normal;
}

void MainWindow::onVerifyerCompleted(Result *result)
{
    m_pending = PendingAction::Verify;
    QObject *p = sender();

    processCompleted(p, result);
}

void MainWindow::onSignerCompleted(Result *result)
{
    m_pending = PendingAction::Sign;
    QObject *p = sender();

    bool success = processCompleted(p, result);

    if (!success) {
        rmAbortedSignature(m_pendingSignature);
        return;
    }

    m_signaturesPending << m_pendingSignature;
    changeSignatureFile(m_pendingSignature);

    m_fprintsPending.removeFirst();

    if (!m_fprintsPending.isEmpty()) {
        sign();
    } else if (m_compress) {
        compressSignature(m_documentPath, m_signaturesPending);
    }
}

void MainWindow::compressSignature(QString doc, QStringList sigs)
{
    QStringList files = QStringList() << doc << sigs;
    QString archive = doc + ".signed.zip";

    if (Zip::compress(archive, files)) {
        for (auto i = files.constBegin();
             i != files.constEnd();
             ++i) {
            QFile::remove(*i);
        }
    }

    QDir dir = QDir(m_dir);
    QString reldoc = dir.relativeFilePath(doc);
    QString relsig;
    if (sigs.count() == 1)
        relsig = dir.relativeFilePath(sigs.at(0));

    changeArchive(archive);

    dir = QDir(m_dir);
    QString absdoc = dir.absoluteFilePath(reldoc);
    changeDocument(absdoc);

    if (sigs.count() == 1)
        changeSignatureFile(dir.absoluteFilePath(relsig));

    m_compress = false;
}

void MainWindow::rmAbortedSignature(QString path)
{
    if (Utils::fileExists(path) && !QFile::remove(path)) {
            error(tr("Error"),
                  tr("Failed to clean up"));
    }
}

void MainWindow::showResult(QString result, Status status)
{
    ui->result->clear();
    if (status == Status::Normal) {
        ui->result->setText(tr("Success"));
    } else {
        ui->result->setText(result);
    }
}

void MainWindow::error(QString title,
                           QString error,
                           QString details,
                           bool stop)
{
    QMessageBox msgbox(this);
    msgbox.setWindowTitle(title);
    msgbox.setText(error);
    msgbox.setDetailedText(details);
    msgbox.exec();

    if (stop)
        stopWaiting();
}

void MainWindow::showDialog()
{
    if (m_dialog)
        m_dialog->exec();
}

void MainWindow::closeDialog()
{
    if (m_dialog)
        m_dialog->close();
    m_dialog.reset();
}

void MainWindow::disableControls()
{
    ui->docView->setEnabled(false);
    ui->certsTable->setEnabled(false);

    ui->sign->setEnabled(false);
    ui->signCompress->setEnabled(false);
    ui->sigChk->setEnabled(false);
}

void MainWindow::enableControls()
{
    ui->docView->setEnabled(true);
    ui->certsTable->setEnabled(true);
    ui->sigChk->setEnabled(true);

    validateSignatureActions();
}

void MainWindow::validateSignatureActions()
{
    bool valid = isValid();
    ui->sign->setEnabled(valid);
    ui->signCompress->setEnabled(valid);
    // We're creating attached signature. No need to pack it into archive
    if (ui->sign->isEnabled()) {
        ui->signCompress->setEnabled(!ui->attached->isChecked());
        if (ui->attached->isChecked()) {
            ui->signCompress->setToolTip(tr("Archive creation is intended to pack detached signatures only."));
        } else {
            ui->signCompress->setToolTip(QString());
        }
    }
}

bool MainWindow::isValid()
{
    if (m_documentPath.isEmpty())
        return false;

    if (ui->certsTable->model()->rowCount() <= 0)
        return false;

    if (ui->certsTable->selectionModel()->selectedRows().count() == 0)
        return false;

    return true;
}

void MainWindow::stopWaiting()
{
    m_waitTimer.stop();
    closeDialog();
    enableControls();
}

void MainWindow::startWaiting(Dialog *dialog)
{
    disableControls();
    m_dialog.reset(dialog);
    m_waitTimer.start();
}

void MainWindow::updateSignatureExtension()
{
    QString path = ui->sigFilename->text();

    if (!path.isEmpty()) {
        QString basename = Utils::splitext(path).at(0);
        QString ext = ui->extension->currentText();

        ui->sigFilename->setText(basename + ext);
    }
}

void MainWindow::generateName()
{
    if (!m_documentPath.isEmpty()) {
        QString name = generateSignatureName(m_documentPath);
        ui->sigFilename->setText(name);
    }
}

QString MainWindow::generateSignatureName(QString path)
{
    QString date = QDateTime::currentDateTime().toString("yy-MM-dd_hh-mm-ss");
    QString base = Utils::splitBasename(path);
    QString ext = ui->extension->currentText();

    return QString("%1_%2%3").arg(date, base, ext);
}

QStringList MainWindow::changeArchive(QString path)
{
    m_archivePath = path;
    m_cache->createDir();
    m_dir = m_cache->cachePath();
    m_documentPath.clear();
    return Zip::decompress(path, m_cache->cachePath());
}

void MainWindow::changeDocument(QString path)
{
    if (path.endsWith(".signed.zip")) {
        QStringList decompressed = changeArchive(path);
        if (decompressed.count() == 2) {
            QString doc;
            QString sig;

            for (auto i = decompressed.constBegin();
                     i != decompressed.constEnd();
                     ++i) {
                QString file = (*i);
                QString extension = Utils::splitextSuffix(file);

                if (currentExtensions().contains(extension)) {
                    sig = file;
                } else {
                    doc = file;
                }
            }

            changeDocument(doc);
            changeSignatureFile(sig);
        } else {
            selectDocument();
        }
        return;
    }

    m_documentPath = path;

    if (path.startsWith(m_cache->cachePath())) {
        path.replace(m_cache->cachePath(), m_archivePath + ":");
    }

    bool valid = !m_documentPath.isEmpty();
    ui->docView->setEnabled(valid);
    ui->generateName->setEnabled(valid);
    ui->sigFilename->setEnabled(valid);
    ui->sign->setEnabled(valid);

    if (!m_documentPath.isEmpty()) {
        ui->sigFilename->setText(generateSignatureName(path));
        ui->docPath->setText(path);
    } else {
        ui->sigFilename->clear();
        ui->docPath->clear();
    }

    changeSignatureFile(QString());
}

void MainWindow::changeSignatureFile(QString path)
{
    m_signaturePath = path;

    if (path.startsWith(m_cache->cachePath())) {
        path.replace(m_cache->cachePath(), m_archivePath + ":");
    }

    QString text;
    if (!path.isEmpty()) {
        text = path;
    } else if (!m_documentPath.isEmpty()) {
        text = tr("Attached signature selected");
    }
    ui->sigPathChk->setText(text);

    validateSignatureActions();
}

void MainWindow::loadCerts()
{
    QList<QStandardItem*> rowData;
    auto model = qobject_cast<QStandardItemModel*>(ui->certsTable->model());
    QList<QHash<QString, QString>> cert_list = Utils::getCerts();
    for (auto i = cert_list.constBegin();
         i != cert_list.constEnd();
         ++i)
    {
        auto certificate = (*i);

        rowData.clear();

        // either "SHA1 Thumbprint" or "SHA1 Hash" could be used
        QString hash;
        if (certificate.contains("SHA1 Thumbprint")) {
                hash = certificate["SHA1 Thumbprint"];
        } else {
                hash = certificate["SHA1 Hash"];
        }

        QString display = certificate["Subject"];
        display += ", " + tr("Fingerprint: ") + hash;

        QString tooltip = certificate["Issuer"];
        tooltip += "\n" + tr("Expires: ") + certificate["Not valid after"];

        auto subject = new QStandardItem(certificate["Subject"]);
        subject->setData(tooltip, Qt::ToolTipRole);
        subject->setData(display, Qt::DisplayRole);
        rowData << subject;

        rowData << new QStandardItem(certificate["Issuer"]);
        rowData << new QStandardItem(hash);

        model->appendRow(rowData);
    }
}

QStringList MainWindow::currentExtensions()
{
    QStringList sigExtensions;

    for (int i = 0; i < ui->extension->count(); i++)
        sigExtensions << ui->extension->itemText(i);

    return sigExtensions;
}

void MainWindow::loadStyleSheet()
{
    QResource::unregisterResource("/etc/alterator/design-browser-qt");
    QResource::registerResource("/etc/alterator/design-browser-qt");

    // set style
    QFile file(":/design/design.qss");
    if (file.exists()) {
        if (file.open(QFile::ReadOnly)) {
            QString styleContent = file.readAll();
            if (styleContent.size() < 10) {
                return;
            }
            qApp->setStyleSheet(styleContent);
        } else {
            return;
        }
    }
}
