/* Copyright (C) 2020-2021 BaseALT Ltd
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <QFileInfo>

#include "process.h"
#include "utils.h"

Result::Result(QProcess::ExitStatus exitStatus,
                                 int exitCode)
{
    m_error = QProcess::ProcessError::UnknownError;
    m_exitStatus = exitStatus;
    m_exitCode = exitCode;
}

Result::Result(QProcess::ProcessError error)
{
    m_error = error;
    m_exitStatus = QProcess::ExitStatus::NormalExit;
    m_exitCode = 0;
}

QProcess::ProcessError Result::error()
{
    return m_error;
}

QProcess::ExitStatus Result::exitStatus()
{
    return m_exitStatus;
}

Process::Process()
{
    m_completedResult = nullptr;
    m_standardOutput = QByteArray();
    m_standardError = QByteArray();

    m_timer.setSingleShot(true);
    connect(&m_timer, &QTimer::timeout, this, &QProcess::kill);
    connect(this, &QProcess::stateChanged, this, &Process::onStateChanged);
    connect(this, QOverload<int, QProcess::ExitStatus>::of(&QProcess::finished),
            this, &Process::onFinished);
    connect(this, &QProcess::errorOccurred, this, &Process::onErrorOccured);
    connect(this, &QProcess::readyReadStandardOutput,
            this, &Process::onReadyReadStandardOutput);
    connect(this, &QProcess::readyReadStandardError,
            this, &Process::onReadyReadStandardError);
}

int Process::killTime()
{
    return m_timer.interval();
}

void Process::setKillTime(int killTime)
{
    m_timer.setInterval(killTime);
}

void Process::readStandardOutput()
{
    QByteArray temp = readAllStandardOutput();
    if (!temp.isEmpty()) {
        m_standardOutput += temp;
    }
}

void Process::readStandardError()
{
    QByteArray temp = readAllStandardError();
    if (!temp.isEmpty()) {
        m_standardError += temp;
    }
}

void Process::onReadyReadStandardOutput()
{
    readStandardOutput();
}

void Process::onReadyReadStandardError()
{
    readStandardError();
}

QString Process::errorCode()
{
    QList<QByteArray> stdout = m_standardOutput.split('\n');
    for (auto i = stdout.constBegin();
         i != stdout.constEnd();
         ++i)
    {
        QString line = (*i);
        QString errorCode = Utils::parseErrorCode(line);
        if (!errorCode.isEmpty())
            return errorCode;
    }
    return QString();
}

bool Process::errorOccured()
{
    QString error = errorCode();
    return (!error.isEmpty() && error != "0x00000000");
}

QByteArray Process::standardOutput()
{
    return m_standardOutput;
}

QByteArray Process::standardError()
{
    return m_standardError;
}

void Process::onFinished(int exitCode, ExitStatus exitStatus)
{
    m_timer.stop();
    m_completedResult = new Result(exitStatus, exitCode);
    emit completed(m_completedResult);
}

void Process::onErrorOccured(ProcessError error)
{
    m_timer.stop();
    if (error == Crashed)
        return;

    m_completedResult = new Result(error);
    emit completed(m_completedResult);
}

void Process::onStateChanged(ProcessState state)
{
    if (state == ProcessState::Starting)
        onStarting();
}

void Process::onStarting()
{
    delete m_completedResult;
    m_completedResult = nullptr;
    m_standardOutput.clear();
    m_standardError.clear();

    if(killTime() > 0)
        m_timer.start();
}

QString Process::command()
{
    return program() + arguments().join(" ");
}

QList<Qt::Key> Process::questionKeys(QByteArray array)
{
    QList<Qt::Key> foundKeys;
    QString text = QString(array).toUtf8();
    QStringList lines = text.split('\n');
    for (auto i = lines.constBegin();
         i != lines.constEnd();
         ++i) {
        QString line = (*i);
        if (line.indexOf("Yes[Y]") || line.indexOf("Да[Y]")) {
            foundKeys.append(Qt::Key_Y);
        }
        if (line.indexOf("No[N]") || line.indexOf("Нет[N]")) {
            foundKeys.append(Qt::Key_N);
        }
        if (line.indexOf("Cancel[C]") || line.indexOf("Отмена[C]")) {
            foundKeys.append(Qt::Key_C);
        }
    }
    return foundKeys;
}

void Process::verify(const QString filePath, const QString sigPath)
{
    setProgram(Paths::binFilePath("cryptcp"));
    setKillTime(0);

    QStringList arguments = QStringList() << "-verify" << "-errchain";
    if (!sigPath.isEmpty()) {
        arguments << "-verall" << "-detached" << filePath << sigPath;
    } else {
        arguments << filePath;
    }

    setArguments(arguments);
    start();
}

void Process::sign(QString path, QString sig, QString fprint, QString derFormat, bool attached)
{
    setProgram(Paths::binFilePath("cryptcp"));
    setKillTime(0);

    QStringList arguments = QStringList() << "-sign" << "-thumbprint" << fprint;
    if (derFormat == "DER")
        arguments << "-der";

    if (attached)
        arguments << path << sig;
    else
        arguments << "-detached" << path << sig;

    setArguments(arguments);
    start();
}

bool Process::findContainers()
{
    setProgram(Paths::binFilePath("csptestf"));
    QStringList arguments = QStringList() << "-absorb" << "-certs" << "-autoprov";
    setArguments(arguments);
    setKillTime(0);

    QProcessEnvironment env = QProcessEnvironment::systemEnvironment();
    env.insert("LC_ALL", "C");
    setProcessEnvironment(env);

    start(QIODevice::ReadOnly);
    waitForFinished();
    return exitCode() == 0 && m_completedResult->exitStatus() == QProcess::NormalExit;
}

bool Process::getCertificates()
{
    setProgram(Paths::binFilePath("certmgr"));
    QStringList arguments = QStringList() << "-list";
    setArguments(arguments);
    setKillTime(0);

    QProcessEnvironment env = QProcessEnvironment::systemEnvironment();
    env.insert("LC_ALL", "C");
    setProcessEnvironment(env);

    start(QIODevice::ReadOnly);
    waitForFinished();
    return exitCode() == 0 && m_completedResult->exitStatus() == QProcess::NormalExit;
}
