/* Copyright (C) 2020-2021 BaseALT Ltd
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <unistd.h>

#include <QDesktopServices>
#include <QDir>
#include <QDirIterator>
#include <QFileInfo>
#include <QMessageBox>
#include <QMimeDatabase>
#include <QStandardPaths>
#include <QUrl>
#include <QRegularExpression>

#include "utils.h"
#include "process.h"

namespace Utils
{
    bool prepareCerts(void)
    {
        Process *systemCheck = new Process();

        if (!systemCheck->findContainers()) {
            qWarning("Failed to find containers <%s>", qPrintable(systemCheck->program()));
            return false;
        }

        QList<QByteArray> out = systemCheck->standardOutput().split('\n');
        for (auto check_it = out.constBegin();
             check_it != out.constEnd();
             ++check_it) {
            if (check_it->startsWith("Match:")) {
                return true;
            }
        }

        return true;
    }

    QList<QHash<QString, QString>> getCerts()
    {
        QList<QHash<QString, QString>> certs;

        Process *certsList = new Process();
        if (!certsList->getCertificates()) {
            return certs;
        }

        QString certificate;
        QStringList certificate_list;

        QList<QByteArray> out = certsList->standardOutput().split('\n');
        for (auto certs_line = out.constBegin();
             certs_line != out.constEnd();
             ++certs_line) {
            QString line = QString(*certs_line);
            if (!line.indexOf(QRegularExpression("\\d+------"))) {
                certificate_list << certificate;
                certificate = QString();
            } else {
                certificate += line.toUtf8() + "\n";
            }
        }
        certificate_list << certificate;

        // Extract "key : value" pairs
        QRegularExpression re("^(?<key>\\w.*?):(?<value>.*?)$",
                              QRegularExpression::MultilineOption |
                              QRegularExpression::UseUnicodePropertiesOption);
        for (QStringList::const_iterator certificate = certificate_list.constBegin();
             certificate != certificate_list.constEnd();
             ++certificate) {
            QHash<QString, QString> cert_fields;
            QRegularExpressionMatchIterator i = re.globalMatch((*certificate));

            while (i.hasNext()) {
                QRegularExpressionMatch match = i.next();
                cert_fields.insert(match.captured("key").trimmed(),
                                   match.captured("value").trimmed());
            }

            // Since 5.0.12000 "SHA1 hash" became "SHA1 Thumbprint"
            if (cert_fields.contains("Issuer") &&
                cert_fields.contains("Subject") &&
                (cert_fields.contains("SHA1 Hash") ||
                 cert_fields.contains("SHA1 Thumbprint"))
                ) {
                certs << cert_fields;
            }
        }

        return certs;
    }

    bool dirExists(QString path)
    {
        QFileInfo info(path);
        return info.isDir();
    }

    bool fileExists(QString path)
    {
        QFileInfo info(path);
        return info.isFile();
    }

    QStringList splitext(QString path)
    {
        return QStringList() << splitextPrefix(path) << splitextSuffix(path);
    }

    QString splitextSuffix(QString path)
    {
        QFileInfo info(path);
        QString extension = "." + info.suffix();
        return extension;
    }

    QString splitextPrefix(QString path)
    {
        QFileInfo info(path);
        QString basename = info.completeBaseName();
        return basename;
    }

    QString splitDirname(QString path)
    {
        QFileInfo info(path);
        return info.dir().path();
    }

    QString splitBasename(QString path)
    {
        QFileInfo info(path);
        return info.fileName();
    }

    bool packageInstalled(QString pkg)
    {
        Process *process = new Process();
        process->start("rpm", QStringList() << "-q" << pkg);
        process->waitForFinished();

        if (process->exitCode() == 0)
            return true;
        return false;
    }

    QString parseErrorCode(QString line)
    {
        line = line.simplified();
        if (line.startsWith("[ErrorCode")) {
            int start = line.indexOf(':');

            if (start < 0)
                return QString();
            else
                start++;

            int end = line.lastIndexOf(']');
            if (end < 0)
                end = line.size() - 1;

            QString errorCode = line.mid(start, end - start);
            errorCode = errorCode.simplified();
            return errorCode;
        }
        return QString();
    }

    QString insertInfix(QString path, QString infix)
    {
        return Utils::splitDirname(path) + "/" + splitextPrefix(path) + "." + infix + splitextSuffix(path);
    }
}
